import sys
if sys.version_info >= (3, 6):
    from pathlib import Path  # pylint: disable=import-error
else:
    from pathlib2 import Path  # pylint: disable=import-error
import pkg_resources
from pyfakefs import fake_filesystem_unittest

import yada.xdg
import yada.config
import yada.repo


class TestRepo(fake_filesystem_unittest.TestCase):
    def setUp(self):
        self.setUpPyfakefs(modules_to_reload=[yada.xdg, yada.config])
        self.fs.add_real_paths([pkg_resources.resource_filename("yada.data", "config.yaml")])
        self.user = yada.config.get_default_user_name()
        self.home = yada.config.get_home()

    def test_repo_relative_to__default(self):
        repo = yada.repo.get_default_repo()
        
        assert repo.path_relative_to(self.home / "vim") == \
            Path("../.local/share/yada/{user}/dot".format(user=self.user))
        assert repo.path_relative_to(self.home / "vim/x") == \
            Path("../../.local/share/yada/{user}/dot".format(user=self.user))
        assert repo.path_relative_to(self.home / "vim/") == \
            Path("../.local/share/yada/{user}/dot".format(user=self.user))
        assert repo.path_relative_to(self.home / "a/b/c/d") == \
            Path("../../../../.local/share/yada/{user}/dot".format(user=self.user))
        assert repo.path_relative_to(self.home / "") == \
            Path(".local/share/yada/{user}/dot".format(user=self.user))
        assert repo.path_relative_to(self.home) == \
            Path(".local/share/yada/{user}/dot".format(user=self.user))

    def test_repo_relative_to__root(self):
        repo = yada.repo.get_default_repo(yada_home="/etc/yada")

        assert repo.path_relative_to(self.home / "vim") == \
            Path("../../../etc/yada/{user}/dot".format(user=self.user))


class TestModule(fake_filesystem_unittest.TestCase):

    def test_module_relative_to(self):
        pass
        #repo = yada.repo.get_default_repo()
        #assert False



import os.path
import sys
if sys.version_info >= (3, 6):
    import pathlib  # pylint: disable=import-error
else:
    import pathlib2 as pathlib  # pylint: disable=import-error
import shlex
if sys.version_info >= (3, 3):
    from shlex import quote  # pylint: disable=import-error
else:
    from pipes import quote
import shutil
import subprocess

import yada.config


def get_default_repo(yada_home=yada.config.get_yada_home()):
    return Repo(yada.config.get_default_user_name(),
                yada.config.get_default_repo_name(),
                yada_home=yada_home)


class Module():
    def __init__(self, repo, name):
        self.repo = repo
        self.name = name

    def __str__(self):
        return self.name

    @property
    def path(self):
        return self.repo.path / "modules" / self.name

    @property
    def files_path(self):
        return self.path / "files"

    @property
    def readme_path(self):
        for ext in [".md", ".txt", ""]:
            path = (self.path / "README{ext}".format(ext=ext))
            if path.exists():
                return path
        return None

    def exists(self):
        return self.path.exists() and self.path.is_dir()

    def create(self):
        self.files_path.mkdir(parents=True, exist_ok=True)

    def import_file(self, path):
        if not path.is_absolute():
            path = pathlib.Path.cwd() / path
        destination = self.files_path / path.relative_to(yada.config.get_home())

        def copy(src, dst):
            try:
                shutil.copy(str(src), str(dst))
            except shutil.SameFileError:
                pass

        return (
            Operation(lambda: destination.parent.mkdir(parents=True, exist_ok=True),
                      "mkdir -p {dst}".format(dst=destination),
                      interactive=False),
            Operation(lambda: copy(path, destination),
                      "cp {src} {dst}".format(src=quote(str(path)),
                                              dst=quote(str(destination))))
        )

    def install(self):
        for f in self.files_path.glob("**/*"):
            destination = yada.config.get_home() / f.relative_to(self.files_path)
            if f.is_dir():
                yield Operation(lambda: destination.mkdir(parents=True, exist_ok=True),
                                "mkdir -p {}".format(quote(str(destination))),
                                interactive=True)
            else:
                command = "ln -sf {src} {dst}".format(
                    src=quote(str(self.repo.path_relative_to(destination.parent) / "modules" / self.name / "files" / f.relative_to(self.files_path))),
                    dst=quote(str(destination)))
                def backup_and_link():
                    if destination.is_file():
                        shutil.copy(str(destination),
                                    str(destination.with_suffix(destination.suffix + '.bkp')))
                    subprocess.call(shlex.split(command), cwd=str(destination.parent))
                yield Operation(backup_and_link,
                                command,
                                interactive=True)

    def push(self, ssh_host):
        for f in self.files_path.glob("**/*"):
            destination = yada.config.get_home() / f.relative_to(self.files_path)
            if f.is_dir():
                command = "ssh {ssh_host} mkdir -p {dir}".format(
                    ssh_host=ssh_host,
                    dir=quote(str(destination)))
                yield Operation(lambda: subprocess.call(shlex.split(command)),
                                command,
                                interactive=True)
            else:
                command = "scp -q {src} {ssh_host}:{dst}".format(
                    ssh_host=ssh_host,
                    src=quote(str(f)),
                    dst=quote(str(destination)))
                def backup_and_link():
                    backup_command = "ssh {ssh_host} {subcommand}".format(
                        ssh_host=ssh_host,
                        subcommand=quote("sh -c \"cp {dst} {dst}.bkp > /dev/null 2>&1\"".format(
                            dst=destination)))
                    subprocess.call(shlex.split(backup_command))
                    subprocess.call(shlex.split(command))
                yield Operation(backup_and_link,
                                command,
                                interactive=True)


class Repo():
    def __init__(self, user, name, yada_home=yada.config.get_yada_home()):
        self.user = user
        self.name = name
        self.yada_home = pathlib.Path(yada_home)

    def __str__(self):
        return "{user}/{name}".format(user=self.user, name=self.name)

    @property
    def path(self):
        return (self.yada_home / self.user / self.name)

    def git_url(self, protocol="ssh"):
        if protocol == "ssh":
            return "git@github.com:{user}/{name}.git".format(user=self.user, name=self.name)
        elif protocol == "https":
            return "https://github.com/{user}/{name}".format(user=self.user, name=self.name)

    def create(self):
        (self.path / "modules").mkdir(parents=True, exist_ok=True)
        subprocess.call(["git", "init"], cwd=str(self.path))

    def exists(self):
        return (self.path / ".git").exists() and (self.path / ".git").is_dir()

    def path_relative_to(self, path):
        common_parent = max(set(list(path.resolve().parents) + [path.resolve()]).intersection(self.path.resolve().parents))
        distance = len(path.parts) - len(common_parent.parts)
        return pathlib.Path(os.path.join(*[".." for _ in range(distance)]) if distance else "") \
            / self.yada_home.relative_to(common_parent) \
            / self.user \
            / self.name

    def module(self, module_name):
        return Module(self, module_name)


class Operation():
    def __init__(self, f, command, interactive=True):
        self.f = f
        self.command = command
        self.interactive = interactive

    def execute(self):
        self.f()



import getpass
import os
import os.path
from pkg_resources import get_distribution
import sys
if sys.version_info >= (3, 6):
    import pathlib  # pylint: disable=import-error
else:
    import pathlib2 as pathlib  # pylint: disable=import-error
import subprocess

import click

import yada.config
import yada.repo
from yada.xdg import XDG_DATA_HOME


class ClickPath(click.Path):
    """
    A Click path argument that returns a `Path`, not a string.
    """

    def convert(self, value, param, ctx):
        return pathlib.Path(super(ClickPath, self).convert(value=value, param=param, ctx=ctx))


class RepoType(click.ParamType):
    name = "Repo"

    def convert(self, value, param, ctx):
        if value is None:
            return None
        elif isinstance(value, yada.repo.Repo):
            return value
        elif "/" in value:
            user, repo = value.split("/")
        else:
            user, repo = yada.config.get_default_user_name(), value
        return yada.repo.Repo(user=user, name=repo, yada_home=ctx.obj["yada-home"])


@click.group()
@click.option("--dry-run/--no-dry-run", default=False,
              help="don't actually make any changes to the filesystem")
@click.option("--yada-home",
              type=ClickPath(file_okay=False, dir_okay=True, writable=True),
              default=str(yada.config.get_yada_home()),
              help="directory to store yada modules in")
@click.pass_context
def cli(ctx, dry_run, yada_home):
    yada_home.mkdir(parents=True, exist_ok=True)

    ctx.ensure_object(dict)
    ctx.obj["dry-run"] = dry_run
    ctx.obj["yada-home"] = yada_home


@cli.command()
def version():
    click.echo(get_distribution("yada").version)


@cli.command()
@click.pass_context
@click.option("--name", type=str, default=yada.config.get_default_repo_name(),
              help="")
def init(ctx, name):
    repo = yada.repo.Repo(user=yada.config.get_default_user_name(), name=name)

    if repo.exists():
        click.secho("Repo {path} already exists.".format(path=repo), fg="yellow")
    else:
        repo.create()


@cli.command()
@click.pass_context
@click.option("--https/--ssh", default=False,
              help="whether to clone via HTTPS or SSH (default)")
@click.option("--repo", "-r", type=RepoType(), default=yada.repo.get_default_repo())
def pull(ctx, https, repo):
    if not repo.exists():
        repo.path.parent.mkdir(parents=True, exist_ok=True)
        subprocess.call(["git", "clone", repo.git_url(protocol="https" if https else "ssh")],
                        cwd=str(repo.path.parent))
    else:
        subprocess.call(["git", "pull"], cwd=str(repo.path))


@cli.command()
@click.pass_context
@click.option("--repo", "-r", type=RepoType(), default=yada.repo.get_default_repo())
def home(ctx, repo):
    click.echo(repo.path)

@cli.command("import")
@click.pass_context
@click.option("--interactive/--no-interactive", "-i", default=False,
              help="query for confirmation before every filesystem operation")
@click.option("--repo", "-r", type=RepoType(), default=yada.repo.get_default_repo())
@click.argument("module", type=str, nargs=1)
@click.argument("files", type=ClickPath(exists=True), nargs=-1)
def import_files(ctx, interactive, repo, module, files):
    module = repo.module(module)
    q = list(files)
    for f in q:
        if f.is_file():
            operations = module.import_file(f)
            for operation in operations:
                click.secho(operation.command, fg="yellow")
                if not ctx.obj["dry-run"] and (not interactive or not operation.interactive or
                                               click.confirm("Take action?")):
                    operation.execute()
        elif f.is_dir():
            q += list(f.glob("**/*"))
        else:
            click.secho("{} is an invalid file to import".format(f), fg="red")


@cli.command("install")
@click.pass_context
@click.option("--interactive/--no-interactive", "-i", default=False,
              help="query for confirmation before every filesystem operation")
@click.option("--repo", "-r", type=RepoType(), default=yada.repo.get_default_repo())
@click.argument("modules", type=str, nargs=-1)
def install(ctx, interactive, repo, modules):
    for module in modules:
        module = repo.module(module)
        click.secho("Installing module {}...".format(module), fg="green")
        for operation in module.install():
            click.secho(operation.command, fg="yellow")
            if not ctx.obj["dry-run"] and (not interactive or not operation.interactive or
                                           click.confirm("Take action?")):
                operation.execute()

    click.secho("Done!", fg="green")


@cli.command("push")
@click.pass_context
@click.option("--interactive/--no-interactive", "-i", default=False,
              help="query for confirmation before each file upload")
@click.option("--repo", "-r", type=RepoType(), default=yada.repo.get_default_repo())
@click.argument("module", type=str)
@click.argument("ssh-host", type=str)
def push(ctx, interactive, repo, module, ssh_host):
    module = repo.module(module)

    for operation in module.push(ssh_host):
        click.secho(operation.command, fg="yellow")
        if not ctx.obj["dry-run"] and (not interactive or not operation.interactive or
                                       click.confirm("Take action?")):
            operation.execute()

    click.secho("Done!", fg="green")


@cli.command("info")
@click.pass_context
@click.option("--repo", "-r", type=RepoType(), default=yada.repo.get_default_repo())
@click.argument("module")
def info(ctx, repo, module):
    module = repo.module(module)

    def header(title, width=80):
        ew = (width - len(title) - 2) // 2
        return "=" * ew + " " + title + " " + "=" * (width - (ew + len(title) + 2))

    if not module.exists():
        click.secho("Module {repo}:{module} not found!".format(repo=repo, module=module), fg="red")
        sys.exit(1)

    click.secho(header("{repo}:{module}".format(repo=repo, module=module)), fg="yellow")

    if module.readme_path:
        click.echo(open(module.readme_path, "r").read())
    else:
        click.secho("(No README found)", fg="bright_black")

    if module.files_path.exists():
        click.secho(header("FILES"), fg="yellow")
        for root, _, files in os.walk(str(module.files_path)):
            level = root.replace(str(module.files_path), '').count(os.sep)
            indent = ' ' * 4 * (level)
            click.secho('{}{}/'.format(indent, os.path.basename(root)), fg="bright_black")
            subindent = ' ' * 4 * (level + 1)
            for f in files:
                target_path = yada.config.get_home() / \
                    pathlib.Path(root, f).relative_to(module.files_path)
                if not target_path.exists():
                    fg = "red"
                elif target_path.is_symlink() and \
                        target_path.resolve() == pathlib.Path(root, f).resolve():
                    fg = "green"
                else:
                    fg = "yellow"
                click.secho('{}{}'.format(subindent, f), fg=fg)

    click.secho(header("RECENT CHANGES"), fg="yellow")
    subprocess.call(["git",
                     "log",
                     "--graph",
                     "--pretty=format:%Cred%h%Creset -%Creset %s %Cgreen(%cr) %C(bold blue)<%an>%Creset",
                     "--abbrev-commit",
                     "--date=relative",
                     "--max-count=10",
                     "--",
                     "modules/{module}".format(module=module)],
                    cwd=str(repo.path), stdout=sys.stdout, stderr=subprocess.STDOUT)



# -*- coding: utf-8 -*-
# Copyright © 2016-2019 Scott Stevenson <scott@stevenson.io>
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all
# copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE
# AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
# DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
# PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

"""XDG Base Directory Specification variables.
XDG_CACHE_HOME, XDG_CONFIG_HOME, and XDG_DATA_HOME are pathlib.Path
objects containing the value of the environment variable of the same
name, or the default defined in the specification if the environment
variable is unset or empty.
XDG_CONFIG_DIRS and XDG_DATA_DIRS are lists of pathlib.Path objects
containing the value of the environment variable of the same name split
on colons, or the default defined in the specification if the
environment variable is unset or empty.
XDG_RUNTIME_DIR is a pathlib.Path object containing the value of the
environment variable of the same name, or None if the environment
variable is not set.
"""

import os
import sys
if sys.version_info >= (3, 6):
    from pathlib import Path  # pylint: disable=import-error
else:
    from pathlib2 import Path  # pylint: disable=import-error


__all__ = [
    "XDG_CACHE_HOME",
    "XDG_CONFIG_DIRS",
    "XDG_CONFIG_HOME",
    "XDG_DATA_DIRS",
    "XDG_DATA_HOME",
    "XDG_RUNTIME_DIR",
]

HOME = Path.home()


def _path_from_env(variable, default):
    """Read an environment variable as a path.
    The environment variable with the specified name is read, and its
    value returned as a path. If the environment variable is not set, or
    set to the empty string, the default value is returned.
    Parameters
    ----------
    variable : str
        Name of the environment variable.
    default : Path
        Default value.
    Returns
    -------
    Path
        Value from environment or default.
    """
    # TODO(srstevenson): Use assignment expression in Python 3.8.
    value = os.environ.get(variable)
    if value:
        return Path(value)
    return default


def _paths_from_env(variable, default):
    """Read an environment variable as a list of paths.
    The environment variable with the specified name is read, and its
    value split on colons and returned as a list of paths. If the
    environment variable is not set, or set to the empty string, the
    default value is returned.
    Parameters
    ----------
    variable : str
        Name of the environment variable.
    default : List[Path]
        Default value.
    Returns
    -------
    List[Path]
        Value from environment or default.
    """
    # TODO(srstevenson): Use assignment expression in Python 3.8.
    value = os.environ.get(variable)
    if value:
        return [Path(path) for path in value.split(":")]
    return default


XDG_CACHE_HOME = _path_from_env("XDG_CACHE_HOME", HOME / ".cache")

XDG_CONFIG_DIRS = _paths_from_env("XDG_CONFIG_DIRS", [Path("/etc/xdg")])

XDG_CONFIG_HOME = _path_from_env("XDG_CONFIG_HOME", HOME / ".config")

XDG_DATA_DIRS = _paths_from_env(
    "XDG_DATA_DIRS",
    [Path(path) for path in "/usr/local/share/:/usr/share/".split(":")],
)

XDG_DATA_HOME = _path_from_env("XDG_DATA_HOME", HOME / ".local" / "share")

try:
    XDG_RUNTIME_DIR = Path(os.environ["XDG_RUNTIME_DIR"])
except KeyError:
    XDG_RUNTIME_DIR = None



from setuptools import setup, find_packages

setup(
    name='yada',
    use_scm_version=True,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/apetresc/yada',
    author='Adrian Petrescu',
    author_email='adrian@apetre.sc',
    description='Yet another dotfile aggregator',

    setup_requires=[
        'pytest-runner==5.1',
        'setuptools_scm==3.3.3'
    ],
    install_requires=[
        'Click>=7.0,<8.0',
        'pathlib2==2.3.3;python_version<"3.4"',
        'pyyaml>=5.1,<6.0'
    ],
    tests_require=[
        'pytest==4.6.3',
        'pyfakefs==3.6',
        'tox==3.12.1'
    ],
    package_data={'yada.data': ['config.yaml']},

    entry_points='''
        [console_scripts]
        yada=yada.cli.main:cli
    '''
)
